const fs = require('fs');
const path = require('path');
const DateHandler = require('../functions/DateHandler');
const SystemLogger = require('./SystemLogger');

class Logger {
  constructor(config = {}) {
    const {
      fileName = 'log',
      filePath = null,
      fileExtension = 'txt',
      useTimeStamp = true
    } = config;
    this.fileExtension = fileExtension;
    this.filePath = filePath ? filePath : this.getDefaultFilePath();
    this.fileName = fileName;
    this.useTimeStamp = useTimeStamp;
  }

  getDefaultFilePath() {
    if(process.env.NODE_ENV === 'development') {
      return process.env.LOG_FOLDER;
    } else {
      return path.join(process.env.LOG_FOLDER, process.argv[2]);
    }
  }

  format(error, info) {
    const sysdate = DateHandler.toString({ useTime: true });
    return `[${sysdate}] ${error}\n${info}\n`;
  }

  async write(error, info = {}, options = {}) {
    const formattedInfo = this.format(error, info);
    const { NODE_ENV = 'development' } = process.env;
    if(NODE_ENV === 'development' && !this.hideConsoleLogger) SystemLogger.error(formattedInfo);
    const { useWriteFile = false, ...fsOptions } = options;
    const fsFunctionToCall = useWriteFile ? fs.writeFile : fs.appendFile;
    return new Promise((resolve, reject) => {
      const path = this.getFullPath();
      const data = formattedInfo;
      const options = fsOptions;
      const callback = (err) => {
        if(err) return reject(err);
        resolve({
          fileName: this.getFormattedFileName(),
        });
      };
      fsFunctionToCall.apply(this, [path, data, options, callback]);
    });
  }

  /**
   * Setters and Getters
   */

  setFileName(fileName) {
    this.fileName = fileName;
  }

  getFileName() {
    return this.fileName;
  }

  getFormattedFileName() {
    let formattedFileName = this.fileName || 'log';
    if(this.useTimeStamp) formattedFileName = `${formattedFileName}_${DateHandler.toString()}`;
    return `${formattedFileName}.${this.fileExtension}`;
  }

  setFilePath(path) {
    this.filePath = path;
  }

  getFilePath() {
    return this.filePath;
  }

  getFullPath() {
    return path.join(this.filePath, this.getFormattedFileName());
  }
}

module.exports = Logger;