const apiController = require('../../controller/ApiController');
const updaterProdutosModel = require('../../model/updaterProdutosModel');
const updaterProdVersaoModel = require('../../model/updaterProdVersaoModel');
const updaterHistoricoModel = require('../../model/updaterHistoricoModel');
const updaterServiceManagementFili = require('../../model/updaterServiceManagementModel');
const updaterModel = require('../../model/updaterModel');
const updaterModelCasher = require('../../model/updaterModelCasher');
const versionCheck = require('../../functions/VersionCheck');
const socketService = require('../../service/SocketService');
const UpdaterWarning = require('../../helpers/UpdaterWarning');
const OracleService = require('../../service/OracleService');

const route = (router) => {
  router.get('/updaterProdutos', (async (req, res) => {
    updaterProdutosModel.getUpdaterProdutos(true).then((products) => {
      let { rows } = products;
      const {
        search,
        search_in
      } = req.query || {};
      if (search && typeof search_in === 'string') {
        rows = rows.filter(row => {
          return (row[search_in] || '').toString().toLowerCase().includes(search);
        });
      }
      const response = {
        data: rows,
        pagination: {
          limit: 999,
          total: rows.length
        }
      };
      res.send(response);
    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.get('/updaterProdutosPerCasher', (async (req, res) => {
    const productId = req.query.PRODUCT_ID;
    updaterProdutosModel.getUpdaterProdutosPerCasherFilial(productId).then((products) => {
      const response = {
        data: products.rows,
        pagination: {
          limit: 999,
          total: products.rows.length
        }
      };
      res.send(response);

    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.get('/updaterProdutosFilial', (async (req, res) => {
    const productId = req.query.PRODUCT_ID;
    if (!productId) {
      res.send([]);
    } else {
      const latestVersion = await apiController.getLatestProductVersion(productId);

      updaterProdutosModel.getUpdaterProdutosWithUnits(productId).then(async (products) => {
        let updaterSockets = await socketService.getUpdaterSocket().getUpdaters();
        const connectedUnits = [];

        await apiController.setLatestUpdaterApprovedVersion();
        const newestUpdateVersion = apiController.getLatestUpdaterApprovedVersion();
        const unitVersions = {};

        updaterSockets.forEach((socket) => {
          if (!socket.CDCAIXA) {
            connectedUnits.push(socket.CDFILIAL);
            if (newestUpdateVersion !== 'error') {
              unitVersions[socket.CDFILIAL] = {};
              unitVersions[socket.CDFILIAL].UPDATE_AVAILABLE = versionCheck.greaterThan(newestUpdateVersion, socket.handshake.headers.version);
              unitVersions[socket.CDFILIAL].CURRENT_VERSION = socket.handshake.headers.version;
            }
          }
        });

        products.rows = products.rows.map((product) => {
          product.CONNECTION_STATUS = connectedUnits.includes(product.CDFILIAL);
          product.LATEST_VERSION = latestVersion;
          product.WARNING_INFO = product.WARNING ? UpdaterWarning.getActiveWarnings({
            unit: product.CDFILIAL,
          }) : [];
          product.WARNING = !!product.WARNING_INFO.length;
          const versionIncompatible = product.WARNING && !product.CONNECTION_STATUS;
          if (newestUpdateVersion !== 'error') {
            const {
              CURRENT_VERSION = null,
              UPDATE_AVAILABLE = null,
            } = unitVersions[product.CDFILIAL] || {};

            product.UPDATE_AVAILABLE = UPDATE_AVAILABLE;
            product.UPDATE_CURRENT_VERSION = CURRENT_VERSION;
            product.UPDATE_AVAILABLE_VERSION = newestUpdateVersion;
            product.UPDATE_AVAILABLE = product.UPDATE_AVAILABLE || versionIncompatible;
          }

          return product;
        });

        const response = {
          data: products.rows,
          pagination: {
            limit: 999,
            total: products.rows.length
          }
        };
        res.send(response);
      }).catch((error) => {
        res.send(error.message || error);
      });
    }
  }));

  router.post('/updaterProductMultiInstall', (async (req, res) => {
    const { PRODUCT_ID, UNITS } = req.body;
    if (!PRODUCT_ID) {
      res.send([]);
      return;
    }
    try {
      const productVersions = await updaterProdutosModel.getUpdaterProductsWithVersion(PRODUCT_ID);
      const updaterSockets = await socketService.getUpdaterSocket().getUpdaters();
      const connectedUnits = [];
      updaterSockets.forEach((socket) => {
        if (!socket.CDCAIXA) {
          connectedUnits.push(socket.CDFILIAL);
        }
      });
      const onlySelectedUnits = ({ CDFILIAL }) => UNITS.includes(CDFILIAL);
      const fillConnectionStatus = row => row.CONNECTION_STATUS = connectedUnits.includes(row.CDFILIAL);
      productVersions.rows = productVersions.rows.filter(onlySelectedUnits);
      productVersions.rows.forEach(fillConnectionStatus);
      const response = {
        data: productVersions.rows,
        pagination: {
          limit: 999,
          total: productVersions.rows.length
        }
      };
      res.send(response);
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  }));

  router.post('/getEventsInProgressByProduct', async (req, res) => {
    const { PRODUCT_ID, UNITS } = req.body;
    if (!PRODUCT_ID) {
      return res.status(400).send('PRODUCT_NOT_INFORMED');
    }
    try {
      const EVENTS_IN_PROGRESS = await updaterModel.getEventsInProgressByProduct({
        PRODUCT_ID,
        CDFILIAL: `_${UNITS.join('_')}_`
      });
      res.send({
        EVENTS: EVENTS_IN_PROGRESS.rows
      });
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  });

  router.get('/updaterProdutosFilialPerCasher', (async (req, res) => {
    updaterProdutosModel.getUpdaterProdutosWithUnitsPerCasher(req.query).then(async (products) => {
      const casherAndProducts = await updaterModelCasher.getCasherWithoutProduct({
        CDFILIAL: req.query.CDFILIAL,
        UPDATERPRODID: req.query.PRODUCT_ID,
      });
      if(casherAndProducts && casherAndProducts.rows && casherAndProducts.rows.length) {
        await Promise.all(casherAndProducts.rows.map((rowToInsert) => {
          return updaterModelCasher.insertUpdaterProdFiliCx(rowToInsert).then(result => result.connection.commit());
        })).catch(err => {
          OracleService.rollback();
        });
        products = await updaterProdutosModel.getUpdaterProdutosWithUnitsPerCasher(req.query);
      }
      let socketUpdaters = await socketService.getUpdaterSocket().getUpdaters();
      const connectedFiliais = socketUpdaters.reduce((acc, socket) => {
        acc[socket.CDFILIAL] = acc[socket.CDFILIAL] || new Set();
        if (socket.CDCAIXA) {
          acc[socket.CDFILIAL].add(socket.CDCAIXA);
        }
        return acc;
      }, {});
      const unitsWithWarning = UpdaterWarning.getMappedUnitsCashersWithWarning();
      products.rows = products.rows.map((product) => {
        const curUnit = connectedFiliais[product.CDFILIAL] || new Set();
        const cashersWithWarning = unitsWithWarning[product.CDFILIAL] || [];
        product.CONNECTION_STATUS = curUnit.has(product.CDCAIXA);
        product.WARNING = cashersWithWarning.includes(product.CDCAIXA);
        return product;
      });
      const response = {
        data: products.rows,
        pagination: {
          limit: 999,
          total: products.rows.length
        }
      };
      res.send(response);
    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.get('/updaterProdutosVersao', (async (req, res) => {
    const productId = req.query.CDPRODUCTID;
    const { search } = req.query || {};
    const { USE_VERSION_FILTER } = process.env;
    if (!productId) {
      res.send([]);
    } else {
      updaterProdVersaoModel.getUpdaterProdVersaoByProduct(productId, USE_VERSION_FILTER).then((products) => {
        let { rows } = products;
        if (search) {
          rows = rows.filter(row => {
            return (row.NRVERSAO || '').includes(search);
          });
        }

        rows.sort((x, y) => versionCheck.sort(x.NRVERSAO, y.NRVERSAO));
        const response = {
          data: rows,
          pagination: {
            limit: 999,
            total: products.rows.length
          }
        };
        res.send(response);

      }).catch((error) => {
        res.send(error.message || error);
      });
    }
  }));

  router.get('/updaterProdutosVersaoConfig', async (req, res) => {
    try {
      const products = await updaterProdVersaoModel.getUpdaterProdVersaoConfig();
  
      const sortedProducts = products.rows.sort((x, y) => {
        return versionCheck.sort(x.NRVERSAO, y.NRVERSAO);
      });
      const response = {
        data: sortedProducts,
        pagination: {
          limit: 999,
          total: sortedProducts.length
        }
      };
      res.send(response);
    } catch (error) {
      res.send(error.message || error);
    }
  });
  
  router.get('/updaterProdutoHistorico', (async (req, res) => {
    const cdProductId = req.query.CDPRODUCTID;
    const cdFilial = req.query.CDFILIAL;
    updaterHistoricoModel.getUpdaterProdutoHistorico(cdProductId, cdFilial).then((historico) => {
      const response = {
        data: historico.rows,
        pagination: {
          limit: 999,
          total: historico.rows.length
        }
      };
      res.send(response);
    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.get('/updaterProdutoHistoricoPorCaixa', (async (req, res) => {
    const cdProductId = req.query.CDPRODUCTID;
    const cdCaixa = req.query.CDCAIXA;
    const cdFilial = req.query.CDFILIAL;
    updaterHistoricoModel.getUpdaterProdutoHistoricoPorCaixa(cdProductId, cdFilial, cdCaixa).then((historico) => {
      const response = {
        data: historico.rows,
        pagination: {
          limit: 999,
          total: historico.rows.length
        }
      };
      res.send(response);
    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.put('/updateUpdaterProdVersaoAprov', (async (req, res) => {
    try {
      const result = await updaterProdVersaoModel.updateUpdaterProdVersaoAprov({
        ID: req.body.ID,
        IDAPROVVER: req.body.IDAPROVVER,
      });

      result.connection.commit();
      res.send('Sucesso!');
    } catch (error) {
      res.send(error.message || error);
    }
  }));

  router.get('/updaterServiceManagementFili', (async (req, res) => {
    const { search } = req.query || {};
    updaterServiceManagementFili.filterServiceManagementFili().then(({ rows }) => {

      if (search) {
        rows = rows.filter(row => {
          return (row.FILIAL || '').includes(search);
        });
      }

      const response = {
        data: rows,
        pagination: {
          limit: 999,
          total: rows.length
        }
      };
      res.send(response);
    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.get('/updaterServiceManagement', (async (req, res) => {
    const cdFilial = req.query.CDFILIAL;

    updaterServiceManagementFili.filterServiceManagementMonitoring(cdFilial).then(({ rows }) => {
      let serviceManagement = [];
      rows.map((element) => {
        const service = element.MONITORAMENTO.split(';');
        service.forEach((item) => {
          serviceManagement.push({
            CDFILIAL: element.CDFILIAL,
            MONITORAMENTO: item
          })
        })
      })

      const response = {
        data: serviceManagement,
        pagination: {
          limit: 999,
          total: serviceManagement.length
        }
      };
      res.send(response);
    }).catch((error) => {
      res.send(error.message || error);
    });
  }));

  router.get('/minimumSupportedVersion', async (req, res) => {
    const { MIN_UPDATER_VERSION_SUPPORTED } = process.env;
    res.send({ minimumSupportedVersionUpdater: MIN_UPDATER_VERSION_SUPPORTED });
  })
}

module.exports = route;