// O valor dos campos de data não considerem o fuso horário local
process.env.ORA_SDTZ = 'UTC';

const oracledb = require('oracledb');
const path = require('path');
const fs = require('fs');
const DBParameter = require('../helpers/DBParameter');
const moment = require('moment');
const DataBaseLogger = require('../logs/DataBaseLogger');
const SystemLogger = require('../logs/SystemLogger');
require('moment/locale/pt-br');

const Now = function () {
  return moment().format('Y-MM-DD HH:mm:ss');
};

class OracleService {
  constructor() {
    this.oracledb = oracledb;
    this.connection = {};
  }

  async connect() {
    try {
      if (await this.checkConnection()) {
        return this.connection;
      }
      
      let connect = await this.getNewConnection();

      this.connection = connect;
      return connect;
    } catch (error) {
      DataBaseLogger.write(error, {type: 'CONNECTION'});
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  async fetch(sqlString, params) {
    try {
      let sqlPrepare = sqlString;

      if (params) {
        Object.keys(params).forEach((key) => {
          const replace = new RegExp(`:${key}`, 'g');
          sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'`);
        });
      }

      if (! await this.checkConnection()) {
        this.connection = await this.connect();

        await this.connection.execute("ALTER SESSION SET TIME_ZONE='UTC'");
      }

      return this.connection.execute(sqlPrepare);
    } catch (error) {
      DataBaseLogger.write(error, {type: 'QUERY_EXECUTION', sql: sqlString, params})
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  async fetchAssoc(sqlString, params, executionParams = {}) {
    try {
      let sqlPrepare = sqlString;
      if (params) {
        Object.keys(params).forEach((key) => {
          const replace = new RegExp(`:${key}`, 'g');
          sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'`);
        });
      }    
    
      if (! await this.checkConnection()) {
        this.connection = await this.connect();

        await this.connection.execute("ALTER SESSION SET TIME_ZONE='UTC'");
      }
      oracledb.fetchAsString = [ oracledb.CLOB ];
      const oracleParams = {
        ...executionParams,
        outFormat: oracledb.OUT_FORMAT_OBJECT
      }
      const result = await this.connection.execute(
        sqlPrepare,
        [],
        oracleParams
      );
      return result;
    } catch (error) {
      DataBaseLogger.write(error, {type: 'QUERY_EXECUTION', sql: sqlString, params})
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  async fetchAssocSingleResult(sqlString, params) {
    return this.fetchAssoc(sqlString, params, { maxRows: 1 });
  }

  async close() {
    await this.connection.close(function (err) {
      if (err) {
        SystemLogger.error(err);
      }
    });
  }

  async commit() {
    if(await this.checkConnection()) {
      await this.connection.commit();
    }
  }

  async rollback() {
    if (await this.checkConnection()) {
      await this.connection.rollback();
    }
  }

  async checkConnection() {
    if(!this.connection) return false;
    if (Object.keys(this.connection).length === 0) {
      return false;
    }

    if (typeof this.connection.ping !== 'function') {
      return false;
    }

    return this.connection.ping().then((err) => {
      if(err) {
        this.connection.close();
        return false;
      }
      return true;
    }).catch( () => {
      return false;
    });
  }

  async execProcedure(name, params) {
    try {
      await this.connect();
      const paramNames = [];
      const procedureParams = params.reduce((procParams, param) => {
        const newParam = DBParameter.create(param);
        procParams[param.name] = newParam;
        paramNames.push(`:${param.name}`);
        return procParams;
      }, {});
      const sql = `BEGIN 
        ${name}(${paramNames.join(', ')}); 
      END;`;
      return await this.connection.execute(sql, procedureParams);
    } catch (error) {
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  getParamType(type) {
    const types = {
      'number': oracledb.NUMBER,
      'string': oracledb.STRING,
      'cursor': oracledb.CURSOR
    }

    return types[type];
  }

  createReturnBinds(outParams, sqlPrepare) {
    const sqlParams = outParams.map((p) => p.name).join(', ');
    const sqlParamsInto = outParams.map((p) => `:${p.name}`).join(', ');
    const sqlReturn = `RETURNING ${sqlParams} INTO ${sqlParamsInto}`;
    const query = sqlPrepare + sqlReturn;

    let bind = {};
    
    for (const param of outParams) {
      bind[param.name] = { type: this.getParamType(param.type), dir: oracledb.BIND_OUT }
    }

    return { query, bind };
  }

  async execute(sqlString, params, outParams = [], connection = null) {
    let sqlPrepare = sqlString;
    let paramsKeys = params ? Object.keys(params) : null;
    let anonymousBlock = false;
    let colPrepare = {};

    paramsKeys = params
      ? paramsKeys.sort(function (a1, a2) {
          return ~(a1.length < a2.length);
        })
      : [];

    if (paramsKeys) {
      paramsKeys.forEach((key) => {
        try {
          const replace = new RegExp(`:${key}(\\s|,|\\)|$)`, 'g');
          const comma = sqlPrepare.match(replace)[0].includes(',') ? ',' : '';
          const closeParenthesis = sqlPrepare.match(replace)[0].includes(')') ? ')' : '';

          if (params[key] == null) {
            sqlPrepare = sqlPrepare.replace(replace, `null${comma}${closeParenthesis}`);
          } else if (typeof params[key] === 'object') {
            anonymousBlock = true;
            const colName = key;

            const regExp = new RegExp('"', 'g');
            let paramData = String.fromCharCode(...params[key].data);
            paramData = paramData.replace(regExp, "'");
            colPrepare[colName] = paramData;
          } else if (typeof params[key] === 'string') {
            const regExp = new RegExp("'", 'g');
            params[key] = params[key].replace(regExp, "''");
            sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'${comma}${closeParenthesis}`);
          } else {
            sqlPrepare = sqlPrepare.replace(replace, `${params[key]}${comma}${closeParenthesis}`);
          }
        } catch (error) {
          // erros que podem acontecer é a chave aparecer no update não como condição e sim para substituir no update
        }
      });
    }
    await this.connect();

    sqlPrepare = sqlPrepare.replace(/TIMESTAMP null/gi, 'null');
    sqlPrepare = sqlPrepare.replace(/TIMESTAMP 'Invalid date'/gi, 'null');

    const dbConnection = connection || this.connection;
    
    if (anonymousBlock) {
      const connect = dbConnection
        .execute(`BEGIN ${sqlPrepare}; END;`, colPrepare)
        .then((result) => {
          const newResult = result;
          newResult.connection = this.connection;

          return newResult;
        })
        .catch((error) => {
          DataBaseLogger.write(error, {type: 'ANONYMOUS_BLOCK', sql: sqlString, params});
          return Promise.reject(new Error('DB_ERROR'));
        });
      return connect;
    }
    try {
      oracledb.fetchAsString = [ oracledb.CLOB ];
      let bindParams = {};

      if (outParams.length) {
        const { query, bind } = this.createReturnBinds(outParams, sqlPrepare);
  
        sqlPrepare = query;
        bindParams = bind;
      }
      
      const dbConnectionResult = await dbConnection.execute(sqlPrepare, bindParams, 
        { 
          outFormat: oracledb.OUT_FORMAT_OBJECT 
        }
      );

      dbConnectionResult.connection = dbConnection;

      return dbConnectionResult;
    } catch (error) {
      DataBaseLogger.write(error, {sql: sqlString, params, type: 'SQL_EXECUTION'});
      this.saveSalesImportException(error, sqlPrepare, params);
      throw error;
    }
  }

  getConnectionParams() {
    const dbName = process.argv[2] || '';
    const { 
      DB_USER: user,
      DB_PASSWORD: password,
      DB_CONN_STR: connectString
     } = process.env;

     if(["EXTERNO", "DEV", "EXTERNAL"].includes(dbName)) {
      return { user, password, connectString };
     }
     return { externalAuth: true, connectString: dbName};
  }

  async getNewConnection() {
    const connParams = this.getConnectionParams();
    return await oracledb.getConnection(connParams);
  }

  async saveSalesImportException(exception, sqlStatement, sqlParams) {
    try {
      let tableName = '';

      const statment = sqlStatement.trim().toUpperCase();
      let testValue = statment.match(/BEGIN/) || statment.match(/DECLARE/);
      if (!testValue) {
        testValue = statment.match(/INTO/) || statment.match(/SET/);
      }

      if (testValue) {
        if (testValue[0] === 'INTO') {
          const inicial = statment.indexOf('INTO') + 5;
          const final = statment.indexOf('(');
          tableName = statment.substring(inicial, final).trim();
        } else {
          const inicial = statment.indexOf('SET');
          tableName = statment.substring(inicial, 6).trim();
        }
      }

      if (tableName !== 'ERROIMPORTAVENDA' && tableName) {
        const primaryKey = `
          SELECT COLUMN_NAME, POSITION FROM ALL_CONS_COLUMNS WHERE CONSTRAINT_NAME = 'PK_${tableName}'  GROUP BY COLUMN_NAME, POSITION ORDER BY POSITION
        `;

        const primaryKeyResult = await this.execute(primaryKey);
        if (tableName === 'VENDA') {
          primaryKeyResult.rows[2][0] = 'NRSEQVENDAMSDE';
        }
        let key = '';

        if (primaryKeyResult.rows) {
          for (const columnName of primaryKeyResult.rows) {
            if (sqlParams[columnName[0]] || sqlParams[`P_${columnName[0]}`]) {
              key =
                `${key + columnName[0]}=${sqlParams[columnName[0]]},` ||
                `${sqlParams[`P_${columnName[0]}`]},`;
            } else {
              key = '';
              break;
            }
          }
        }

        const params = [];
        const properties = Object.keys(sqlParams);

        for (const i of properties) {
          const value = `${i} = ${sqlParams[i]}`;
          params.push(value);
        }

        const dsParams = params.join();
        if (!key) {
          key = dsParams.substring(0, 4000);
        } else {
          key = key.substring(0, key.length - 1).trim();
        }

        const statmentSql = sqlStatement.substring(0, 4000);
        const insertStatement = `
          INSERT INTO ERROIMPORTAVENDA
          (CHAVE, TABELA, DSINSTRUCAOSQL, DSPARAMS, DSERRO, DTERRO)
          VALUES
          (:P_CHAVE, :P_TABELA, SUBSTR(:P_DSINSTRUCAOSQL, 1, 2000), SUBSTR(:P_DSPARAMS, 1, 2000),:P_DSERRO, TIMESTAMP :P_DTERRO)
        `;

        const errorParams = {
          P_CHAVE: key,
          P_TABELA: tableName,
          P_DSINSTRUCAOSQL: statmentSql.trim(),
          P_DSPARAMS: dsParams.substring(0, 4000),
          P_DSERRO: exception['message'],
          P_DTERRO: Now(),
        };

        //Função comentada para funcionar na base local.
        //await this.execute(insertStatement, errorParams);
      }
    } catch (error) {
      SystemLogger.error(error);
    }
  }
}

module.exports = new OracleService();