const io = require("socket.io");
const AuthService = require("./AuthService");
const UpService = require("./UpService");
const UpdaterSocket = require("./UpdaterSocket");
const MonitoringSocket = require("./MonitoringSocket");
const SocketRooms = require("./SocketRooms");
const SystemLogger = require("../logs/SystemLogger");
const MonitoringApacheCloud = require("./MonitoringApacheCloud");
const UpdaterWarning = require("../helpers/UpdaterWarning");

class SocketService {
    getUpdaterSocket() {
        return UpdaterSocket;
    }
    getMonitoringSocket() {
        return MonitoringSocket;
    }
    getMonitoringApacheCloud() {
        return MonitoringApacheCloud;
    }
    async start(http) {
        try {
            this.io = io(http, { cors: "*" });

            UpdaterSocket.setSocketIo(this.io);
            UpdaterSocket.setMonitoringSocket(MonitoringSocket);
            MonitoringSocket.setSocketIo(this.io);
            MonitoringApacheCloud.setSocketIo(this.io);

            this.io.on("connection", (socket) => {
                socket.on('wsConnectionTest', () => {
                    socket.emit('wsConnection', {});
                });
                try {
                    const { log: info, agent, version, cdfilial, cdcaixa, ip, hostname } = this.getSocketHeaderInfo(socket);
                    const { page } = this.getQueryParams(socket);
                    switch (agent) {
                        case "updater":
                            socket.CDFILIAL = cdfilial;
                            socket.CDCAIXA = cdcaixa || null;
                            UpdaterSocket.validateVersion(socket, version, cdfilial, cdcaixa);
                            UpdaterSocket.validateOtherConnections(socket).then(_ => {
                                UpdaterSocket.joinDefaultRoom(socket);
                                SystemLogger.log(`OnConnection (${info}): ${socket.id}`);
                                UpdaterSocket.connectedUpdater({ unit: cdfilial, casher: cdcaixa, connected: true, version });
                            }).catch(error => {
                                const duplicateInfo = {};
                                duplicateInfo[hostname] = ip;
                                UpdaterWarning.addWarning({
                                    unit: cdfilial,
                                    casher: cdcaixa,
                                }, 'DUPLICATE', duplicateInfo);
                                socket.disconnect();
                                SystemLogger.error(error);
                            });
                            break;
                        case "monitor":
                            socket.CDFILIAL = cdfilial;
                            MonitoringSocket.hasConsumersForMonitoring().then( hasConsumers => {
                                if(hasConsumers) {
                                    MonitoringSocket.startMonitoring([socket]);
                                    MonitoringSocket.connectedClient([cdfilial]);
                                }
                            });
                            MonitoringSocket.joinDefaultRoom(socket);
                            SystemLogger.log(`OnConnection (${info}): ${socket.id}`);
                            break;
                        default:
                            switch( page ) {
                                case 'monitoring':
                                    MonitoringSocket.joinRooms(socket);
                                    MonitoringSocket.connectedUser();
                                    MonitoringSocket.startMonitoring();
                                    break;
                                case 'updates':
                                    UpdaterSocket.joinRooms(socket);
                                    break;
                                case 'updatesPerCasher':
                                    UpdaterSocket.joinRooms(socket, true);
                                    break;
                                case 'parametrizacoes':
                                    socket.join(SocketRooms.master.parametrizacoesPage);
                                    break;
                                case 'monitoringCloud':
                                    socket.join(SocketRooms.master.monitoringCloudPage);
                                    MonitoringApacheCloud.startMonitoring();
                                    break;
                                case 'serviceManagement':
                                    socket.join(SocketRooms.master.serviceManagement)
                            }
                    }

                    /** Garante que o portal está conectado */

                    socket.on("login", (data) => {
                        if(!data) {
                            socket.disconnect();
                            return;
                        }
                        this.login(socket, data);
                    });

                    /**
                     * Exportação de vendas
                     */
                    socket.on("iniciaExportacao", (data) => {
                        try {
                            data.payload = JSON.parse(data.payload);
                            UpService.setInsereServidor(
                                data.CDFILIAL,
                                data.CDCAIXA,
                                data.CDLOJA,
                                data.payload
                            );
                        } catch (e) {}
                    });

                    socket.on("disconnect", async () => {
                        const { agent, cdfilial, cdcaixa } = this.getSocketHeaderInfo(socket);
                        switch(agent.toLowerCase()) {
                            case 'updater':
                                const socketUpdater = await UpdaterSocket.getUpdaterSocketByUnitAndCasher(cdfilial, cdcaixa);
                                UpdaterSocket.connectedUpdater({
                                    unit: cdfilial,
                                    casher: cdcaixa,
                                    connected: !!socketUpdater,
                                    version: socket.version,
                                });
                                break;

                            case 'monitor':
                                MonitoringSocket.connectedClient([cdfilial], 'N');
                                break;

                            default: // assistant portal
                                MonitoringSocket.disconnectedUser();
                        }
                    });
                    MonitoringApacheCloud.registerEvents(socket, agent);
                    MonitoringSocket.registerEvents(socket, agent);
                    UpdaterSocket.registerEvents(socket, agent);
                } catch (error) {
                    socket.disconnect();
                    SystemLogger.error(error);
                }
            });

            await MonitoringSocket.loadInitialParams();
            await MonitoringApacheCloud.loadMonitoringParams();
        } catch (error) {
            SystemLogger.error(error);
        }
    }

    getQueryParams(socket) {
        const defaults = {
            page: '',
        };
        const { query } = socket.handshake;
        return ({
            ...defaults,
            ...query
        });
    }

    getSocketHeaderInfo(socket) {
        let { headers } = socket.handshake;
        const defaults = {
            agent: "",
            cdfilial: "",
            cdcaixa: "",
            hostname: "",
            version: "",
            "user-agent": "",
            "x-real-ip": socket.handshake.address,
        };
        let { agent, cdfilial, cdcaixa, hostname, version, ...others } = {
            ...defaults,
            ...headers,
        };
        let ip = others["x-real-ip"];
        let userAgent = others["user-agent"];
        let log = [agent, cdfilial, cdcaixa, hostname, ip]
            .filter((value) => value)
            .join(" | ");
        return { log, agent, cdfilial, cdcaixa, hostname, version, ip, userAgent };
    }

    async login(socket, data) {
        const { agent } = this.getSocketHeaderInfo(socket);
        try {
            const authResult = await AuthService.autenticar(data.email, data.password, socket.CDFILIAL);
            switch(agent) {
                case 'updater':
                    const { CDFILIAL, CDCAIXA = null } = socket;
                    const { UPDATE_AVAILABLE_VERSION } = process.env;
                    const { version } = this.getSocketHeaderInfo(socket);
                    const params = { CDFILIAL, UPDATE_CURRENT_VERSION: version, UPDATE_AVAILABLE_VERSION };
                    if (CDCAIXA) {
                        params.CAIXAS = [CDCAIXA];
                    }
                    break;
                case 'monitor':
                    MonitoringSocket.connectedClient([socket.CDFILIAL]);
                    break;
                default:
                    socket.emit("confirmLogin", {
                        error: false,
                        message: "CONNECTED_SUCCESS",
                        result: authResult,
                    });
                    break;
            }
        } catch (error) {
            SystemLogger.error(error);
            socket.emit("message", "Você não é um usuário autorizado!");
            socket.disconnect();
        }
    }
}

module.exports = new SocketService();