const updaterModel = require("../model/updaterModel");
const base64 = require("../functions/Base64");
const versionCheck = require("../functions/VersionCheck");
const SocketRooms = require("./SocketRooms");
const UpdaterService = require('./UpdaterService');
const UpdaterTelegramData = require('./UpdaterTelegramData');
const SystemLogger = require("../logs/SystemLogger");
const UpdaterLogger = require("../logs/UpdaterLogger");
const AssistantControl = require("./AssistantControl");
const Telegram = require("../functions/Telegram");
const UpdaterWarning = require('../helpers/UpdaterWarning');

class UpdaterSocket {
    monitoringSocket = null;
    setSocketIo(io) {
        this.io = io;
    }
    getSocketIo() {
        return this.io;
    }
    setMonitoringSocket(monitoringSocket) {
        this.monitoringSocket = monitoringSocket;
    }
    getMonitoringSocket() {
        return this.monitoringSocket;
    }
    getUpdateRoomName(CDCAIXA = null) {
        const {
            updatesPage,
            updatesPerCasherPage
        } = SocketRooms.master;
        return CDCAIXA ? updatesPerCasherPage : updatesPage;
    }

    async connectedUpdater({ unit, casher = null, version = '', connected }) {
        const roomName = this.getUpdateRoomName(casher);
        let room = this.getRoom(roomName);
        if(!connected) {
            UpdaterWarning.clearWarning({ unit, casher });
        }
        const updateAvaibleVersion = AssistantControl.getUpdateAvailableVersion();
        const { DATA, ...warningStatus } = UpdaterWarning.getWarning({ unit, casher }) || {};
        const activeWarnings = Object.keys(warningStatus).filter(name => warningStatus[name]);
        const socketData = {
            CONNECTED: connected,
            CDFILIAL: unit,
            WARNING: !!activeWarnings.length,
            VERSION: version,
            WARNING_INFO: activeWarnings,
            UPDATE_AVAILABLE: updateAvaibleVersion && connected && updateAvaibleVersion !== version,
            UPDATE_AVAILABLE_VERSION: updateAvaibleVersion || '',
        }
        if(casher) {
            socketData.CDCAIXA = casher;
        } else {
            room = room.to(SocketRooms.master.parametrizacoesPage);
        }
        room.emit('updaterConnected', socketData);
    }
    joinDefaultRoom( socket ) {
        socket.join(SocketRooms.updater.default);
    }
    joinRooms( socket, isPerCasher = false ) {
        const pages = [SocketRooms.master.default, this.getUpdateRoomName(isPerCasher)];
        socket.join(pages);
    }
    async hasUnitUpdaterConnected(CDFILIAL, CDCAIXA = null) {
        const socketUpdater = await this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);
        return !!socketUpdater;
    }
    async validateOtherConnections(socket) {
        const { CDFILIAL, CDCAIXA } = socket;
        const socketUpdater = await this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);
        const unitConnected = !!socketUpdater && socketUpdater.id !== socket.id;
        if(unitConnected) {
            const { headers = {}, address} = socketUpdater.handshake;
            const { hostname } = headers;
            UpdaterWarning.addWarning({ unit: CDFILIAL, casher: CDCAIXA }, 'DUPLICATE', {
                ACTIVE: { hostname, address }
            });
            const room = this.getUpdateRoomName(CDCAIXA);
            const socketData = {
                CDFILIAL,
                WARNING: true,
                WARNING_INFO: ['DUPLICATE'],
            }
            if (CDCAIXA) {
                socketData.CDCAIXA = CDCAIXA;
            }
            this.getRoom(room).emit('updaterWarning', socketData);
            let casher = CDCAIXA ? `_${CDCAIXA}` : '';
            let errorMessage = `UNIT_${CDFILIAL}${casher}_IS_ALREADY_CONNECTED`;
            return Promise.reject(new Error(errorMessage));
        }
        return Promise.resolve();
    }
    validateVersion(socket, version, CDFILIAL, CDCAIXA) {
        const { MIN_UPDATER_VERSION_SUPPORTED = 'any' } = process.env;
        if (MIN_UPDATER_VERSION_SUPPORTED !== 'any' && versionCheck.greaterThan(MIN_UPDATER_VERSION_SUPPORTED, version)) {
            socket.emit("invalidVersion", {
                supportedUpdaterVersion: MIN_UPDATER_VERSION_SUPPORTED,
                updaterVersion: version,
            });
            UpdaterWarning.addWarning({ unit: CDFILIAL, casher: CDCAIXA }, 'INVALID_VERSION', {
                CURRENT_VERSION: version,
                MIN_VERSION: MIN_UPDATER_VERSION_SUPPORTED,
            })
            return false;
        }
        UpdaterWarning.removeWarning({ unit: CDFILIAL, casher: CDCAIXA }, 'INVALID_VERSION');
        socket.emit("validVersion");
        return true;
    }
    getRoom(roomName) {
        return this.getSocketIo().to(roomName);
    }
    getDownloadLogRoom(CDFILIAL, CDCAIXA, type) {
        const {
            downloadLogTxt,
            downloadLogZip
        } = SocketRooms.master;
        const roomFunction = type === 'txt' ? downloadLogTxt : downloadLogZip;
        const downloadLogRoom = roomFunction.apply(this, [CDFILIAL, CDCAIXA]);
        return downloadLogRoom;
    }
    async getUpdaters() {
        return await this.getRoom(SocketRooms.updater.default).fetchSockets();
    }
    async getMasterFrontendClients(id = null){
        const sockets = await this.getRoom(SocketRooms.master.default).fetchSockets();
        return id ? sockets.find( sckt => sckt.id === id) : sockets;
    }
    async getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA = null) {
        const socketUpdaters = await this.getUpdaters();
        const matchSocket = (socket) => {
            return socket.CDFILIAL == CDFILIAL && (!CDCAIXA || socket.CDCAIXA === CDCAIXA);
        };
        const updaterUnitSocket = socketUpdaters.find( matchSocket );
        return updaterUnitSocket;
    }
    registerUpdaterSocketEvents(socket) {
        socket.on("updateConfigResult", (resultConfigIni) => {

            this.getRoom(SocketRooms.master.parametrizacoesPage).emit(
                "updateConfigIniFinish",
                resultConfigIni
            );
        });

        socket.on("updaterLogResponse", async (logs) => {
            if(!logs.CDFILIAL) return;
            const {
                masterSocketId,
                usingLatestLines = false,
                ...responseLogs
            } = logs;
            const roomLogsUnit = this.getRoom(SocketRooms.updater.log(socket.CDFILIAL, socket.CDCAIXA));
            if(usingLatestLines) {
                const masterSockets = await this.getMasterFrontendClients();
                const targetSocket = masterSockets.find( socket => socket.id === masterSocketId);
                if(targetSocket) targetSocket.emit('updaterLog', responseLogs);
            } else {
                roomLogsUnit.emit('updaterLog', responseLogs);
            }
        });

        socket.on("selfUpdateResponseSuccess", async (result) => {
            const room = this.getUpdateRoomName(socket.CDCAIXA);
            this.getRoom(room).emit("selfUpdateSuccess", result);
        });

        socket.on("selfUpdateResponseError", async (result) => {
            const room = this.getUpdateRoomName(socket.CDCAIXA);
            this.getRoom(room).emit("selfUpdateError", result);
        });

        socket.on("updaterResultInProgress", async (result) => {
            try {
                const room = this.getUpdateRoomName(socket.CDCAIXA);
                const paramsRoom = this.getRoom(SocketRooms.master.parametrizacoesPage);
                if(result.action !== 'verify') {
                    this.getRoom(room).emit("updaterInProgress", result);
                }
                switch (result.action) {
                    case 'install':
                        result.payload.OPERATION = 'INSTALACAO_EM_CURSO';
                        break;
                    case 'uninstall':
                        result.payload.OPERATION = 'REMOCAO_EM_CURSO';
                        break;
                    case 'verify':
                        result.payload.OPERATION = 'VERIFICACAO_EM_CURSO';
                        paramsRoom.emit('updaterInProgress', result);
                        break;
                    case 'integrity_analyze':
                        result.payload.OPERATION = 'ANALISE_INTEGRIDADE_EM_CURSO';
                        break;
                }

                await UpdaterService.updateEventTable(result.payload);
            } catch (err) {
                SystemLogger.error(err);
            }
        });

        socket.on("updaterResultFinish", async (result) => {
            try {
                const paramsRoom = SocketRooms.master.parametrizacoesPage;
                const roomName = this.getUpdateRoomName(socket.CDCAIXA);
                const room = this.getRoom(roomName);
                const { CDFILIAL } = socket;
                const socketResponse = { ...result, CDFILIAL };
                const { operation } = socketResponse;
                switch(operation) {
                    case 'VERIFY':
                        if(socketResponse.error) {
                            room.to(paramsRoom).emit('updaterFinish', socketResponse);
                        } else {
                            UpdaterService.handleInstalledProducts(socketResponse.payload).then((response) => {
                                if(response.invalidProducts) {
                                    socketResponse.INVALID_PRODUCTS = response.invalidProducts;
                                }
                                this.getRoom(paramsRoom).emit('updaterFinish', socketResponse);
                            }).catch(err => {
                                this.getRoom(paramsRoom).emit('updaterFinish', {
                                    ...socketResponse,
                                    error: true,
                                });
                            });
                        }
                        break;
                    case 'INTEGRITY_ANALYZE':
                        UpdaterService.updateAnalyzeIntegrity(socketResponse.payload).then(() => {
                            room.emit("updaterFinish", socketResponse);
                        }).catch(err => {
                            room.emit("updaterFinish", { ... socketResponse, error: true });
                            SystemLogger.error(err);
                        });
                        break;
                    case 'INSTALL':
                    case 'UPDATE':
                    case 'REMOVE':
                        room.emit("updaterFinish", socketResponse);
                        const DSHASHANALISE = socketResponse.payload.hash || null;
                        const PRODFILIID = socketResponse.payload.PRODFILIID;
                        if(operation === 'REMOVE') {
                            await UpdaterService.updateResetIntegrityAnalyze(PRODFILIID);
                        } else {
                            await UpdaterService.updateIntegrityHash({ DSHASHANALISE, PRODFILIID });
                        }
                        break;
                }
            } catch(err) {
                SystemLogger.error(err);
            }
        });

        socket.on("updaterChangeServices", (result) => {
            const { CDFILIAL } = socket;
            switch(result.operation.toUpperCase()) {
                case 'INSTALL':
                    this.monitoringSocket.addServiceToMonitoring(CDFILIAL, result.services);
                    UpdaterService.handleChangeServiceByUnit(CDFILIAL, result.services, 'INSTALL');
                    break;
                case 'REMOVE':
                    this.monitoringSocket.removeServiceToMonitoring(CDFILIAL, result.services);
                    UpdaterService.handleChangeServiceByUnit(CDFILIAL, result.services, 'REMOVE');
                    break;
                default:
                    SystemLogger.error(`OPERATION_NOT_MAPPED ${result.operation}`);
            }
        })
        socket.on("testConnectionResult", (resultTestConn) => {
            this.getRoom(SocketRooms.master.parametrizacoesPage).emit(
                "testConnectionFinish",
                resultTestConn
            );
        });

        socket.on('downloadLogError', ({ error, message, param }) => {
            const { CDFILIAL, CDCAIXA, type } = param;
            const downloadLogRoomName = this.getDownloadLogRoom(CDFILIAL, CDCAIXA, type);
            const room = this.getRoom(downloadLogRoomName);
            room && room.emit('downloadLogResult', { error, message, param });
        });

        socket.on('serviceManagementResult', (param) => {
            const updaterServiceManagementRoom = this.getRoom(SocketRooms.master.serviceManagement);
            if (updaterServiceManagementRoom) {
                updaterServiceManagementRoom.emit('serviceManagementResult', param);
            }
        });
    }

    registerMasterSocketEvents(socket) {
        socket.on("updateConfigInit", async (params) => {
            const {
                CDFILIAL,
                CONFIG
            } = params;
            const socketUpdater = await this.getUpdaterSocketByUnitAndCasher( CDFILIAL );
            if (socketUpdater) {
                try {
                    socketUpdater.emit("updateConfig", CONFIG);
                } catch (error) {
                    socket.emit("updateConfigIniFinish", { error: true, message: error.message });
                }
            } else {
                socket.emit(
                    "updateConfigIniFinish",
                    {
                        error: true,
                        message: "CONNECTION_ACTIVE_REQUIRED_UPDATE_CONFIG_FILE_UNIT",
                    }
                );
            }
        });

        socket.on("getUpdaterLog", async (params) => {
            const updaterUnitSocket = await this.getUpdaterSocketByUnitAndCasher(
                params.CDFILIAL, params.CDCAIXA
            );
            if (updaterUnitSocket) {
                socket.join(SocketRooms.updater.log(params.CDFILIAL, params.CDCAIXA));
                updaterUnitSocket.emit("getUpdaterLog", {...params, masterSocketId: socket.id});
            }
        });

        socket.on('stopUpdaterLog', async (params) => {
            if(!params.CDFILIAL) return;
            const roomName = SocketRooms.updater.log(params.CDFILIAL, params.CDCAIXA);
            socket.leave(roomName);
            const roomLogs = this.getRoom(roomName);
            const connectedSockets = await roomLogs.fetchSockets();
            if(!connectedSockets.length) {
                const updaterSocket = await this.getUpdaterSocketByUnitAndCasher(params.CDFILIAL);
                if(updaterSocket) updaterSocket.emit('unwatchLogFile', params);
            }
        });

        socket.on('leaveDownloadLogRoom', async(params) => {
            const { CDFILIAL, CDCAIXA, type } = params;
            const roomName = this.getDownloadLogRoom(CDFILIAL, CDCAIXA, type);
            socket.leave(roomName);
        });

        socket.on("execUpdater", async (params) => {
            const updaterSocketUnit = await this.getUpdaterSocketByUnitAndCasher(params.CDFILIAL, params.CDCAIXA);
            if (updaterSocketUnit) {
                // Envia requisição de atualização para a loja
                updaterSocketUnit.emit("updater", params);
            } else {
                SystemLogger.error(`SOCKET_UPDATER_NOT_FOUND_${params.CDFILIAL}`);
            }
        });

        socket.on('execUpdaterMultiInstall', async (params) => {
            const socketUpdaters = await this.getUpdaters();
            const unitsInfo = params.UNITS.reduce((acc, unitInfo) => {
                const {CDFILIAL, ...info} = unitInfo;
                acc[CDFILIAL] = info;
                return acc;
            }, {});
            socketUpdaters.forEach(socket => {
                const isTargetSocket = (socket.CDFILIAL in unitsInfo) && !socket.CDCAIXA;
                if(isTargetSocket) {
                    const { PRODFILIID, CURRENTVERSION } = unitsInfo[socket.CDFILIAL];
                    socket.emit('updater', {
                        ...params,
                        PRODFILIID,
                        CDFILIAL: socket.CDFILIAL,
                        CDPRODUCTID: params.DEVOPSID.toString(),
                        DEVOPSID: params.DEVOPSID.toString(),
                        CURRENTVERSION: CURRENTVERSION === '-' ? null : CURRENTVERSION
                    });
                }
            })
        });

        socket.on('testConnection', async (CDFILIAL) => {
            const socketUpdater = await this.getUpdaterSocketByUnitAndCasher( CDFILIAL );

            if(socketUpdater) {
                try {
                    const { rows } = await updaterModel.getConfigIniByUnit({P_CDFILIAL: CDFILIAL});

                    if (rows.length === 0) {
                        throw new Error(`Não foi encontrado o registro do arquivo de configuração para a filial ${CDFILIAL}`);
                    }
                    const configIniDecode = base64.decode(rows[0].CONFIG);

                    socketUpdater.emit("updaterTestConnection", configIniDecode);
                } catch(error) {
                    socket.emit("testConnectionFinish", { error: true, message: error.message });
                }
            }
            else {
                socket.emit("testConnectionFinish", { error: true, message: "Falha no teste de conexão. Não existe conexão ativa com a loja no momento." });
            }
        });

        socket.on('detailStatusConnection', async({CDFILIAL, CDCAIXA = null}) => {
            const { DATA: { DUPLICATE = {} } = {} } = UpdaterWarning.getState({ unit: CDFILIAL, casher: CDCAIXA }) || {};
            socket.emit('detailStatusConnectionResult', DUPLICATE);
        });

        socket.on('selfUpdateRequest', async (param) => {
            const { LATEST, NMOPERADOR, CDFILIAL, NRORG, CDCAIXA = null } = param;
            const updaterSocket = await this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);

            if (updaterSocket) {
                const basicAuth = AssistantControl.getBasicAuth();
                const controlUrl = AssistantControl.getControlUrl();
                const data = {
                    ip: process.env.NODE_ENV === 'development' ? 'Ambiente de Desenvolvimento' : socket.handshake.address,
                    user: NMOPERADOR,
                    CDFILIAL,
                    NRORG,
                }
                UpdaterTelegramData.getData(data, 'self-updater').then(telegramData => {
                    Telegram.sendNotification('telegram-update-system', telegramData);
                });
                // Envia requisição de atualização para a loja
                updaterSocket.emit("selfUpdate", { OPERATOR: NMOPERADOR, NEWVERSION: LATEST, URL: `${controlUrl}/updater-version/download`, AUTH_TOKEN: `Basic ${basicAuth}` });
            } else {
                SystemLogger.error(`SOCKET_UPDATER_NOT_FOUND_${CDFILIAL}`);
            }
        });

        socket.on('downloadLog', async (param) => {
            try {
                const { CDFILIAL, CDCAIXA = null, PRODUCT, type } = param;

                const { fileName, useLocalLog } = await UpdaterLogger.getLocalLogInfo(CDFILIAL, CDCAIXA, PRODUCT);

                if (useLocalLog && type === 'zip') {
                    const { UPDATER_LOG_SERVER } = process.env;

                    param.downloadUrl = `${UPDATER_LOG_SERVER}/${fileName}`;
                    socket.emit('downloadLogResult', {error: false, param});
                    return;
                }

                const updaterSocket = await this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);

                if (updaterSocket) {
                    const downloadLogRoom = this.getDownloadLogRoom(CDFILIAL, CDCAIXA, type);

                    socket.join(downloadLogRoom);
                    updaterSocket.emit('downloadLog', param);
                    return;
                }
            } catch (error) {
                SystemLogger.error(error);
            }
        });

        socket.on("serviceManagement", async (params) => {
            const socketUpdater = await this.getUpdaterSocketByUnitAndCasher(params.cdFilial, params.cdCaixa);
            if (socketUpdater) {
                socketUpdater.emit("serviceManagement", params);
            } else {
                socket.emit("serviceManagementResult", {
                    error: true,
                    message: 'CONNECTION_ACTIVE_REQUIRED_UNIT'
                });
            }
        });

        socket.on('removeWarning', async ({ CDFILIAL, CDCAIXA, WARNING_INFO }) => {
            UpdaterWarning.removeWarning({ unit: CDFILIAL, casher: CDCAIXA }, WARNING_INFO);
            const roomName = this.getUpdateRoomName(socket.CDCAIXA);
            const room = this.getRoom(roomName);
            const socketUpdater = this.getUpdaterSocketByUnitAndCasher(CDFILIAL, CDCAIXA);
            const latestVersion = AssistantControl.getUpdateAvailableVersion();
            const warnings = UpdaterWarning.getActiveWarnings({ unit: CDFILIAL, casher: CDCAIXA });
            const connected = !!socketUpdater;
            room.emit('updaterConnected', {
                CDFILIAL,
                CDCAIXA,
                CONNECTED: connected,
                WARNING: !!warnings.length,
                WARNING_INFO: warnings,
                UPDATE_AVAILABLE: connected && latestVersion && socketUpdater.version !== latestVersion,
                UPDATE_AVAILABLE_VERSION: latestVersion,
            });
        });
    }

    registerEvents(socket, agent) {
        if(agent === 'updater') {
            this.registerUpdaterSocketEvents(socket);
        } else if(agent !== 'monitor') {
            this.registerMasterSocketEvents(socket);
        }
    }
}

module.exports = new UpdaterSocket();